import 'package:collection/collection.dart';

abstract class Feature {

  final String name;
  final bool isPart;
  bool get isNotPart => !isPart;
  const Feature._(this.name, this.isPart);

  factory Feature.create(String name, List<String> options) {
    switch(name) {
      case 'console': return const ConsoleFeature._();
      case 'persist': {
        final version = int.tryParse(options.firstOrNull ?? '1') ?? 1;
        return PersistFeature._(version);
      }
      case 'rest': return const RestFeature._();
      case 'trigger': return const TriggerFeature._();
      default:
        print('unknown feature: $name');
        return const UnknownFeature._();
    }
  }

  factory Feature.trigger() => Feature.create('trigger', []);
}

class ConsoleFeature extends Feature {
  const ConsoleFeature._() : super._('console', false);
}
class PersistFeature extends Feature {
  final int version;
  const PersistFeature._(this.version) : super._('persist', true);
}
class RestFeature extends Feature {
  const RestFeature._() : super._('rest', true);
}
class TriggerFeature extends Feature {
  const TriggerFeature._() : super._('trigger', false);
}
class UnknownFeature extends Feature {
  const UnknownFeature._() : super._('unknown', false);
}
