import 'package:oobium_datastore_gen/src/generators/rest_feature_adapter.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';
import 'package:xstring/xstring.dart';

String genRest(Schema schema) => RestFeatureGenerator(schema).compile();

class RestFeatureGenerator {

  final Schema schema;
  Iterable<Model> get models => schema.models;
  RestFeatureGenerator(this.schema);

  String compileCreator() =>
'''
  _RestFeature(
    ${models.map((m) => '${m.adapterVarName}: ${m.adapterVarName} ?? ${m.adapterName}(),').join()}
  )
''';

  String compile() =>
'''
${schema.partOfDeclaration}

class _RestFeature extends BufferedStateFeature {

  ${models.map((m) => 'final ${m.adapterName} ${m.adapterVarName};').join()}
  _RestFeature({
    ${models.map((m) => 'required this.${m.adapterVarName},').join()}
  }) {
    ${models.where((m) => m.isNotInternal).map((m) => '${m.adapterVarName}._feature = this;').join()}
  }
  
  @override
  Future<void> onClose() {
    _reset();
    return super.onClose();
  }
  
  @override
  Future<void> onDestroy() {
    _reset();
    return super.onDestroy();
  }
  
  @override
  void onEvents(StateEvents events) {
    ${['created', 'updated', 'removed'].map((type) => '''
      for(final event in events.${type.varName}(this)) {${models.where((m) => (type != 'removed' || m.isNotSingleton)).map((m) => '''
        if(event is StateEvent<${m.name}>) {
          _log.debug(() => '$type-${m.name}');
          if(${m.adapterVarName}.on${type.titleized} != null) {
            q + (ex) => ${m.adapterVarName}._$type(ex, ${(type == 'updated') ? 'event.oldValue, event.newValue' : 'event.value'});
          }
        }''').join(' else ')}
        else {
          _log.warn(() => '$type-UnknownType: \$event');
        }
      }
    ''').join()}
  }
  
  void _reset() {
    ${models.where((m) => m.canFetch).map((m) => '${m.adapterVarName}.reset();').join()}
  }
}
${models.map((m) => RestAdapterGenerator(m).compile()).join('\n')}
''';
}

extension ModelX on Model {
  String get adapterName => '${name}RestAdapter';
  String get adapterVarName => '${varName}RestAdapter';
}