import 'package:oobium_datastore_gen/src/generators/rest_feature.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';

class RestAdapterGenerator {

  final Model m;
  Schema get schema => m.schema;
  RestAdapterGenerator(this.m);

  String compile() =>
m.isSingleton ?
'''
class ${m.adapterName} {
  late final _RestFeature _feature;
  FutureOr<Map<String, dynamic>?> Function($onRequestParams)? onRequest;
  FutureOr<void> Function($onResultsParams)? onResults;
  FutureOr<void> Function($onChangeParams)? onCreated;
  FutureOr<void> Function($onUpdateParams)? onUpdated;
  Future<void>? _fetched;
  ${m.adapterName}({
    this.onRequest,
    this.onResults,
    this.onCreated,
    this.onUpdated
  });
  FutureOr<void> _created(Executor ex, $_onChangeParams) => onCreated?.call($onChangeArgs);
  FutureOr<void> _updated(Executor ex, $_onUpdateParams) => onUpdated?.call($onUpdateArgs);
  void reset() => _fetched = null;
  Future<void> fetch($fetchParams) {
    if (force) {
      final current = _fetched;
      if(current != null) {
        _fetched = null;
        return current.then((_) => fetch($fetchArgs));
      }
    }
    return (_fetched ??= _fetch($_fetchArgs));
  }
  Future<void> _fetch($_fetchParams) async {
    _log.debug(() => 'fetching ${m.name}');
    final request = onRequest;
    if (request == null) {
      _log.debug(() => 'fetched ${m.name}: skipped, onRequest not set');
    } else {
      final data = await request($onRequestArgs);
      if(data == null) {
        _log.debug(() => 'fetched ${m.name}: not found');
      } else {
        await onResults?.call($onResultsArgs);
        final ${m.varName} = ${m.name}.fromJson($jsonArgs);
        _feature.put(${m.varName});
        _log.debug(() => 'fetched \$${m.varName}');
      }
    }
  }
}
''' :
m.isExternal ?
'''
class ${m.adapterName} {
  late final _RestFeature _feature;
  FutureOr<dynamic> Function($onRequestParams)? onRequest;
  FutureOr<void> Function($onResultsParams)? onResults;
  FutureOr<void> Function($onChangeParams)? onCreated;
  FutureOr<void> Function($onChangeParams)? onRemoved;
  FutureOr<void> Function($onUpdateParams)? onUpdated;
  final _fetched = <int, Future<void>>{};
  ${m.adapterName}({
    this.onRequest,
    this.onResults,
    this.onCreated,
    this.onRemoved,
    this.onUpdated
  });
  FutureOr<void> _created(Executor ex, $_onChangeParams) => onCreated?.call($onChangeArgs);
  FutureOr<void> _removed(Executor ex, $_onChangeParams) => onRemoved?.call($onChangeArgs);
  FutureOr<void> _updated(Executor ex, $_onUpdateParams) => onUpdated?.call($onUpdateArgs);
  void reset() => _fetched.clear();
  Future<void> fetch($fetchParams) {
    if (force) {
      final current = _fetched.remove(query.hashCode);
      if(current != null) {
        return current.then((_) => fetch($fetchArgs));
      }
    }
    return _fetched.putIfAbsent(query.hashCode, () => _fetch($_fetchArgs));
  }
  Future<void> _fetch($_fetchParams) async {
    _log.debug(() => 'fetching ${m.name}(\$query)');
    final request = onRequest;
    if (request == null) {
      _log.debug(() => 'fetched ${m.name}: skipped, onRequest not set');
    } else {
      final data = await request($onRequestArgs);
      if(data == null) {
        _log.debug(() => 'fetched ${m.name}: not found');
      }
      else if(data is Map<String, dynamic>) {
        await onResults?.call($onResultsArgs);
        final ${m.varName} = ${m.name}.fromJson($jsonArgs);
        _feature.put(${m.varName});
        _log.debug(() => 'fetched \$${m.varName}');
      }
      else if(data is Iterable) {
        await onResults?.call($onResultsArgs);
        final ${m.varNamePlural} = ${m.name}.fromJsonList($jsonArgs);
        _feature.set<${m.name}>($setArgs);
        _log.debug(() => 'fetched \$${m.varNamePlural}');
      }
      else {
        _log.debug(() => 'fetched ${m.name}: invalid result(\$data)');
      }
    }
  }
}
''' :
'''
class ${m.adapterName} {
  late final _RestFeature _feature;
  FutureOr<void> Function($onChangeParams)? onCreated;
  FutureOr<void> Function($onChangeParams)? onRemoved;
  FutureOr<void> Function($onUpdateParams)? onUpdated;
  ${m.adapterName}({
    this.onCreated,
    this.onRemoved,
    this.onUpdated
  });
  FutureOr<void> _created(Executor ex, $_onChangeParams) => onCreated?.call($onChangeArgs);
  FutureOr<void> _removed(Executor ex, $_onChangeParams) => onRemoved?.call($onChangeArgs);
  FutureOr<void> _updated(Executor ex, $_onUpdateParams) => onUpdated?.call($onUpdateArgs);
}
''';

  String get onRequestParams => [
    'StateReader state',
    if(m.isNotSingleton) '${m.queryName} query'
  ].join(',');

  String get onRequestArgs => [
    '_feature.asReader()',
    if(m.isNotSingleton) 'query'
  ].join(',');

  String get onResultsParams => [
    'StateWriter state',
    if(m.isNotSingleton) '${m.queryName} query',
    if(m.isSingleton) 'Map<String, dynamic> result'
    else 'dynamic results'
  ].join(',');

  String get onResultsArgs => [
    '_feature.asWriter()',
    if(m.isNotSingleton) 'query',
    'data'
  ].join(',');

  String get onChangeParams => [
    'StateReader state',
    '${m.name} ${m.varName}'
  ].join(',');

  String get _onChangeParams => [
    '${m.name} ${m.varName}'
  ].join(',');

  String get onChangeArgs => [
    '_feature.asReader()',
    m.varName
  ].join(',');

  String get onUpdateParams => [
    'StateReader state',
    '${m.name} old${m.name}',
    '${m.name} ${m.varName}'
  ].join(',');

  String get _onUpdateParams => [
    '${m.name} old${m.name}',
    '${m.name} ${m.varName}'
  ].join(',');

  String get onUpdateArgs => [
    '_feature.asReader()',
    'old${m.name}',
    m.varName
  ].join(',');

  String get fetchParams => [
    if(m.isNotSingleton) '${m.queryName} query',
    'bool force'
  ].join(',');

  String get fetchArgs => [
    if(m.isNotSingleton) 'query',
    'false'
  ].join(',');

  String get _fetchParams => [
    if(m.isNotSingleton) '${m.queryName} query',
  ].join(',');

  String get _fetchArgs => [
    if(m.isNotSingleton) 'query',
  ].join(',');

  String get jsonArgs => [
    'data',
    if(m.isNotSingleton) 'withQuery: query'
  ].join(',');

  String get setArgs => [
    m.varNamePlural,
    'where: query.test',
  ].join(',');
}

