import 'package:oobium_datastore_gen/src/feature.dart';
import 'package:oobium_datastore_gen/src/model.dart';
import 'package:oobium_datastore_gen/src/schema_parser.dart';

class TriggerFeatureGenerator {

  final Schema schema;
  TriggerFeatureGenerator(this.schema);

  String get dsName => schema.dsName;

  bool get hasDependents => models.isNotEmpty;
  Iterable<Model> get models => schema.types.where((m) => m.dependents.any((d) => d.isNotSingleton));

  String compile() =>
schema.has<TriggerFeature>() ?
'''
class _TriggerFeature extends StateFeature {
  @override
  void onEvent(StateEvent e) {
    if(e.isRemove) {
      _onRemove(e);
    }
  }
  
  void _onRemove(StateEvent e) {${models.map((m) => '''
    if(e is StateEvent<${m.name}>) {
      _log.trace(() => 'removed-${m.name}');
      ${remove(m)}
    }''').join(' else ')}
  }
}
''' : '';

  String remove(Model m) => m.dependentsFields.where((f) => f.model.isNotSingleton).map((f) =>
    f.isHasMany
      ? 'e.source.remove(${f.model.keyType}(e.value.${f.name}));'
      : 'for(final ${f.model.varName} in getAll<${f.model.name}>((s) => s.${f.name} == e.value.id)) {'
        '  e.source.remove(${f.model.varName});'
        '}'
  ).join();
}
